// <<--------------------------------------------->>
// << SLADE (SlayeR's 'LeetAss Doom Editor)       >>
// << By Simon Judd, 2004                         >>
// << ------------------------------------------- >>
// << <3dmode.cpp>                                >>
// << General 3d mode stuff                       >>
// <<--------------------------------------------->>

// Includes ------------------------------------- >>
#include "main.h"
#include "struct_3d.h"
#include "camera.h"
#include "map.h"
#include "tex_browse.h"
#include "tex_align.h"

// VARIABLES ----------------------------- >>
int h_line;
//int c_sec;
int h_thing;

string	copy_ftex = "-";
string	copy_wtex = "-";
int		copy_ttype = 1;

CVAR(Bool, render_determine_hilight, true, 0);

float grav = 0.01f;

// EXTERNAL VARIABLES -------------------- >>
extern float scale;
extern rect_3d_t *selected_rect;
extern int selected_floor, selected_ceil;
extern wallinfo_t *walls;
extern ssectinfo_t **sects;
extern sectinfo_t *sectors;
extern Camera camera;
extern Map map;
extern DWORD n_gl_ssectors;
extern bool browser;
extern TextureList tex_list;
extern string msg_text;
extern int msg_count;
extern bool gravity, mix_tex;
extern numlist_t visible_subsectors;
extern gl_vertex *gl_verts;
extern gl_seg *gl_segs;
extern gl_ssect *gl_ssectors;
extern gl_node *gl_nodes;
extern numlist_t aligned_lines;
extern numlist_t sprite_order;
extern sprite_3d_t *sprites3d;

EXTERN_CVAR(Bool, render_things)

// cross: Returns the cross product of 2 vectors
// ------------------------------------------ >>
point3_t cross(point3_t p1, point3_t p2)
{
	point3_t cross_product;	

	cross_product.x = ((p1.y * p2.z) - (p1.z * p2.y));
	cross_product.y = ((p1.z * p2.x) - (p1.x * p2.z));
	cross_product.z = ((p1.x * p2.y) - (p1.y * p2.x));

	return cross_product;
}

// dot_product: Returns the dot product of 2 vectors
// ---------------------------------------------- >>
float dot_product(point3_t v1, point3_t v2)
{
	return (v1.x * v2.x) + (v1.y * v1.y) + (v1.z * v2.z);
}

// angle_between_vectors: Returns the angle between 2 vectors
// ------------------------------------------------------- >>
float angle_between_vectors(float x1, float y1, float x2, float y2)
{
	float dot = (x1 * x2) + (y1 * y2);
	float len = sqrt(x1 * x1 + y1 * y1) * sqrt(x2 * x2 + y2 * y2);
	float cosin = dot / len;

	return acos(cosin) * 180.0f / 3.1415926535897932384f;
}

// line_intersect: Determines if the camera is looking at a line
// (borrowed from Doom Builder, thanks CodeImp :D)
// ---------------------------------------------------------- >>
float line_intersect(point3_t r1, point3_t r2, float x1, float y1, float x2, float y2)
{
	// Calculate the intersection distance from the ray
	float u_ray = ((x2 - x1) * (r1.y - y1) - (y2 - y1) * (r1.x - x1)) /
				  ((y2 - y1) * (r2.x - r1.x) - (x2 - x1) * (r2.y - r1.y));
	
	// Calculate the intersection distance from the line
	float u_line = ((r2.x - r1.x) * (r1.y - y1) - (r2.y - r1.y) * (r1.x - x1)) /
				   ((y2 - y1) * (r2.x - r1.x) - (x2 - x1) * (r2.y - r1.y));
	
	// Return the distance on the ray if intersecting, or return -1
	if((u_ray >= 0) && (u_ray <= 1024) && (u_line >= 0) && (u_line <= 1)) return u_ray; else return -1;
}

// apply_gravity: Applies gravity to the camera
// ----------------------------------------- >>
void apply_gravity()
{
	int sector = determine_sector((int)(camera.position.x / scale), (int)(camera.position.y / scale));

	if (sector != -1)
	{
		float floor_height = 0.0f;
		bool moved = false;

		if (map.sectors[sector]->f_height + 40 < map.sectors[sector]->c_height)
			floor_height = (float)(map.sectors[sector]->f_height + 40) * scale;
		else
		{
			floor_height = (float)(map.sectors[sector]->c_height - 4) * scale;
			float diff = floor_height - camera.position.z;
			camera.position.z = floor_height;
			camera.view.z += diff;
			grav = 0.01f;
			return;
		}

		if (camera.position.z > floor_height + 0.05f)
		{
			grav = grav * 2;
			camera.position.z -= grav;
			camera.view.z -= grav;
			moved = true;
		}

		if (camera.position.z < floor_height - 0.05f)
		{
			grav = grav * 2;
			camera.position.z += grav;
			camera.view.z += grav;
			moved = true;
		}

		if (camera.position.z > floor_height - grav && camera.position.z < floor_height + grav && moved)
		{
			float diff = floor_height - camera.position.z;
			camera.position.z = floor_height;
			camera.view.z += diff;
			grav = 0.01f;
		}
	}
}

// determine_seg_side: Determines the side of a seg a point is on
// ----------------------------------------------------------- >>
bool determine_seg_side(DWORD s, float x, float y)
{
	// Get x1,y1,x2,y2 of seg
	float x1 = 0.0f;
	float y1 = 0.0f;
	float x2 = 0.0f;
	float y2 = 0.0f;
	long v = gl_segs[s].vertex1;

	if (v & SEG_GLVERTEX)
	{
		x1 = gl_verts[v & ~SEG_GLVERTEX].x * scale;
		y1 = gl_verts[v & ~SEG_GLVERTEX].y * scale;
	}
	else
	{
		x1 = map.verts[v]->x * scale;
		y1 = map.verts[v]->y * scale;
	}

	v = gl_segs[s].vertex2;

	if (v & SEG_GLVERTEX)
	{
		x2 = gl_verts[v & ~SEG_GLVERTEX].x * scale;
		y2 = gl_verts[v & ~SEG_GLVERTEX].y * scale;
	}
	else
	{
		x2 = map.verts[v]->x * scale;
		y2 = map.verts[v]->y * scale;
	}

	if (determine_line_side(x1, y1, x2, y2, x, y))
		return true;
	else
		return false;
}

// determine_3d_hilight: Determines what the 3d mode camera is looking at
// ------------------------------------------------------------------- >>
void determine_3d_hilight()
{
	if (!render_determine_hilight)
		return;

	int count = 0;
	float min_dist = 1024;
	point3_t direction = camera.view - camera.position;

	h_line = -1;
	selected_floor = -1;
	selected_ceil = -1;
	selected_rect = NULL;

	numlist_t sectors_check;

	if (render_things)
	{
		for (DWORD a = 0; a < sprite_order.n_numbers; a++)
		{
			int t = sprite_order.numbers[a];
			rect_3d_t rect;
			point3_t p_vector;
			p_vector = cross(camera.view - camera.position, camera.up_vector);
			p_vector = p_vector.normalize();
			int f_height = 0;

			if (sprites3d[t].sector < map.n_sectors)
			{
				if (sprites3d[t].hanging)
					f_height = map.sectors[sprites3d[t].sector]->c_height - sprites3d[t].height;
				else
					f_height = map.sectors[sprites3d[t].sector]->f_height;
			}

			if (map.things[t]->z != 0 && map.hexen)
				f_height = map.things[t]->z;

			if (sprites3d[t].width == -1)
			{
				rect.tl.set((map.things[t]->x - p_vector.x * 8) * scale,
							(map.things[t]->y - p_vector.y * 8) * scale,
							(f_height + 16) * scale);
				rect.br.set((map.things[t]->x + p_vector.x * 8) * scale,
							(map.things[t]->y + p_vector.y * 8) * scale,
							f_height * scale);
			}
			else
			{
				rect.tl.set((map.things[t]->x - p_vector.x * (sprites3d[t].width / 2)) * scale,
							(map.things[t]->y - p_vector.y * (sprites3d[t].width / 2)) * scale,
							(f_height + sprites3d[t].height) * scale);
				rect.br.set((map.things[t]->x + p_vector.x * (sprites3d[t].width / 2)) * scale,
							(map.things[t]->y + p_vector.y * (sprites3d[t].width / 2)) * scale,
							f_height * scale);
			}

			float dist = line_intersect(camera.position, camera.view, rect.tl.x, rect.tl.y, rect.br.x, rect.br.y);

			if (dist != -1 && dist < min_dist)
			{
				point3_t hit_point(camera.position.x + (direction.x * dist),
					camera.position.y + (direction.y * dist),
					camera.position.z + (direction.z * dist));

				if (hit_point.z >= rect.br.z && hit_point.z <= rect.tl.z)
				{
					h_thing = t;
					min_dist = dist;
					selected_rect = NULL;
					selected_ceil = -1;
					selected_floor = -1;
					h_line = -1;
				}
			}
		}
	}

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (walls[l].visible)
		{
			rect_t lrect = map.l_getrect(l);
			float dist = line_intersect(camera.position, camera.view,
				(float)lrect.x1() * scale, (float)lrect.y1() * scale,
				(float)lrect.x2() * scale, (float)lrect.y2() * scale);

			if (dist != -1 && dist < min_dist)
			{

				point3_t hit_point(camera.position.x + (direction.x * dist),
					camera.position.y + (direction.y * dist),
					camera.position.z + (direction.z * dist));

				int side = 0;

				if (determine_line_side(l, camera.position.x / scale, camera.position.y / scale))
					side = 0;
				else
					side = 1;

				// Determine which part of a line we are pointing to (if any)
				if (walls[l].has_mid[side])
				{
					if (hit_point.z >= walls[l].mid[side].br.z && hit_point.z <= walls[l].mid[side].tl.z)
					{
						selected_rect = &walls[l].mid[side];
						h_line = l;
						h_thing = -1;
						min_dist = dist;
						selected_ceil = -1;
						selected_floor = -1;
					}
				}

				if (walls[l].has_up[side])
				{
					if (hit_point.z >= walls[l].up[side].br.z && hit_point.z <= walls[l].up[side].tl.z)
					{
						selected_rect = &walls[l].up[side];
						h_line = l;
						h_thing = -1;
						min_dist = dist;
						selected_ceil = -1;
						selected_floor = -1;
					}
				}

				if (walls[l].has_lo[side])
				{
					if (hit_point.z >= walls[l].lo[side].br.z && hit_point.z <= walls[l].lo[side].tl.z)
					{
						selected_rect = &walls[l].lo[side];
						h_line = l;
						h_thing = -1;
						min_dist = dist;
						selected_ceil = -1;
						selected_floor = -1;
					}
				}
			}
		}
	}

	for (DWORD a = 0; a < visible_subsectors.n_numbers; a++)
	{
		short s = visible_subsectors.numbers[a];

		if (direction.z > 0)
		{
			float height = map.sectors[sects[s]->parent_sector->sector]->c_height * scale;

			if (camera.position.z < height)
			{
				float dist = (height - camera.position.z) / (camera.view.z - camera.position.z);

				if (dist < min_dist)
				{
					point3_t intersection(camera.position.x + (direction.x * dist),
						camera.position.y + (direction.y * dist),
						camera.position.z + (direction.z * dist));
					
					bool in = true;
					DWORD start = gl_ssectors[s].startseg;
					DWORD end = start + gl_ssectors[s].n_segs;
					for (DWORD seg = start; seg < end; seg++)
					{
						if (!determine_seg_side(seg, intersection.x, intersection.y))
						{
							in = false;
							break;
						}
					}

					if (in)
					{
						selected_ceil = sects[s]->parent_sector->sector;
						selected_floor = -1;
						selected_rect = NULL;
						h_line = -1;
						h_thing = -1;
						min_dist = dist;
					}
				}
			}
		}

		if (direction.z < 0)
		{
			float height = map.sectors[sects[s]->parent_sector->sector]->f_height * scale;

			if (camera.position.z > height)
			{
				float dist = (height - camera.position.z) / (camera.view.z - camera.position.z);

				if (dist < min_dist)
				{
					point3_t intersection(camera.position.x + (direction.x * dist),
						camera.position.y + (direction.y * dist),
						camera.position.z + (direction.z * dist));
					
					bool in = true;
					DWORD start = gl_ssectors[s].startseg;
					DWORD end = start + gl_ssectors[s].n_segs;
					for (DWORD seg = start; seg < end; seg++)
					{
						if (!determine_seg_side(seg, intersection.x, intersection.y))
						{
							in = false;
							break;
						}
					}

					if (in)
					{
						selected_floor = sects[s]->parent_sector->sector;
						selected_ceil = -1;
						selected_rect = NULL;
						h_line = -1;
						h_thing = -1;
						min_dist = dist;
					}
				}
			}
		}
	}
}

// update_sector: Updates all ssectors/lines that are in a sector
// ----------------------------------------------------------- >>
void update_sector(int sector)
{
	setup_sector(sector);

	for (DWORD s = 0; s < n_gl_ssectors; s++)
	{
		if (sects[s]->parent_sector->sector == sector)
			setup_ssect(s);
	}

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (map.l_getsector1(l) == sector || map.l_getsector2(l) == sector)
		{
			walls[l].has_mid[0] = false;
			walls[l].has_mid[1] = false;
			walls[l].has_up[0] = false;
			walls[l].has_up[1] = false;
			walls[l].has_lo[0] = false;
			walls[l].has_lo[1] = false;

			if (map.lines[l]->side1 != -1)
			{
				if (map.l_needsmidtex(l, 1))
				{
					setup_wall_rect(l, 1, 0);
					setup_wall_tex(l, 1, 0);
				}

				if (map.l_needsuptex(l, 1))
				{
					setup_wall_rect(l, 1, 1);
					setup_wall_tex(l, 1, 1);
				}

				if (map.l_needslotex(l, 1))
				{
					setup_wall_rect(l, 1, 2);
					setup_wall_tex(l, 1, 2);
				}
			}

			if (map.lines[l]->side2 != -1)
			{
				if (map.l_needsmidtex(l, 2))
				{
					setup_wall_rect(l, 2, 0);
					setup_wall_tex(l, 2, 0);
				}

				if (map.l_needsuptex(l, 2))
				{
					setup_wall_rect(l, 2, 1);
					setup_wall_tex(l, 2, 1);
				}

				if (map.l_needslotex(l, 2))
				{
					setup_wall_rect(l, 2, 2);
					setup_wall_tex(l, 2, 2);
				}
			}
		}
	}

	determine_3d_hilight();
}

// sector_3d_changeheight: Changes floor/ceiling height of hilighted sector
// --------------------------------------------------------------------- >>
void sector_3d_changeheight(bool floor, int amount)
{
	if (h_line == -1)
	{
		if (selected_floor != -1)
		{
			map.s_changeheight(selected_floor, true, amount);
			char temp[6] = "";
			msg_text = "Height: ";
			sprintf(temp, "%d", map.sectors[selected_floor]->f_height);
			msg_text += temp;
			msg_count = 100;
			update_sector(selected_floor);
		}

		if (selected_ceil != -1)
		{
			map.s_changeheight(selected_ceil, false, amount);
			char temp[6] = "";
			msg_text = "Height: ";
			sprintf(temp, "%d", map.sectors[selected_ceil]->c_height);
			msg_text += temp;
			msg_count = 100;
			update_sector(selected_ceil);
		}
	}
	else
	{
		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
		{
			map.s_changeheight(map.l_getsector1(h_line), floor, amount);
			msg_count = 100;

			if (floor)
			{
				char temp[6] = "";
				msg_text = "Floor Height: ";
				sprintf(temp, "%d", map.sectors[map.l_getsector1(h_line)]->f_height);
				msg_text += temp;
			}
			else
			{
				char temp[6] = "";
				msg_text = "Ceiling Height: ";
				sprintf(temp, "%d", map.sectors[map.l_getsector1(h_line)]->c_height);
				msg_text += temp;
			}

			update_sector(map.l_getsector1(h_line));
		}
		else
		{
			map.s_changeheight(map.l_getsector2(h_line), floor, amount);
			msg_count = 100;

			if (floor)
			{
				char temp[6] = "";
				msg_text = "Floor Height: ";
				sprintf(temp, "%d", map.sectors[map.l_getsector2(h_line)]->f_height);
				msg_text += temp;
			}
			else
			{
				char temp[6] = "";
				msg_text = "Ceiling Height: ";
				sprintf(temp, "%d", map.sectors[map.l_getsector2(h_line)]->c_height);
				msg_text += temp;
			}

			update_sector(map.l_getsector2(h_line));
		}
	}
}

// init_3d_browser: Sets up the texture browser depending on what is hilighted
// ------------------------------------------------------------------------ >>
void init_3d_browser()
{
	browser = true;

	if (selected_rect == NULL)
	{
		if (selected_floor != -1)
		{
			if (mix_tex)
				setup_browser(true, true, map.sectors[selected_floor]->f_tex);
			else
				setup_browser(false, true, map.sectors[selected_floor]->f_tex);
			
			//browser_findtex(map.sectors[selected_floor]->f_tex);
		}

		if (selected_ceil != -1)
		{
			if (mix_tex)
				setup_browser(true, true, map.sectors[selected_ceil]->c_tex);
			else
				setup_browser(false, true, map.sectors[selected_ceil]->c_tex);

			//browser_findtex(map.sectors[selected_ceil]->c_tex);
		}

		if (h_thing != -1)
			setup_browser_things();
	}
	else if (h_line != -1)
	{
		if (mix_tex)
				setup_browser(true, true, tex_list.get_name(selected_rect->tex));
			else
				setup_browser(true, false, tex_list.get_name(selected_rect->tex));

		//browser_findtex(tex_list.get_name(selected_rect->tex));
	}
	else
		browser = false;
}

// copy_3d_texture: Copies the hilighted texture
// ------------------------------------------ >>
void copy_3d_texture()
{
	if (selected_rect == NULL)
	{
		if (selected_floor != -1)
		{
			if (mix_tex)
				copy_wtex = map.sectors[selected_floor]->f_tex;

			copy_ftex = map.sectors[selected_floor]->f_tex;
			msg_text = "Copied Texture \"";
			msg_text += copy_ftex;
			msg_text += "\"";
			msg_count = 100;
		}

		if (selected_ceil != -1)
		{
			if (mix_tex)
				copy_wtex = map.sectors[selected_ceil]->c_tex;

			copy_ftex = map.sectors[selected_ceil]->c_tex;
			msg_text = "Copied Texture \"";
			msg_text += copy_ftex;
			msg_text += "\"";
			msg_count = 100;
		}

		if (h_thing != -1)
		{
			copy_ttype = map.things[h_thing]->type;
			msg_text = "Copied Thing \"";
			msg_text += get_tdef(copy_ttype)->name;
			msg_text += "\"";
			msg_count = 100;
		}
	}
	else
	{
		if (mix_tex)
			copy_ftex = tex_list.get_name(selected_rect->tex);

		copy_wtex = tex_list.get_name(selected_rect->tex);
		msg_text = "Copied Texture \"";
		msg_text += copy_wtex;
		msg_text += "\"";
		msg_count = 100;
	}
}

// update_selected_rect: Updates the hilighted wall rect
// -------------------------------------------------- >>
void update_selected_rect()
{
	int l = h_line;

	walls[l].has_mid[0] = false;
	walls[l].has_mid[1] = false;
	walls[l].has_up[0] = false;
	walls[l].has_up[1] = false;
	walls[l].has_lo[0] = false;
	walls[l].has_lo[1] = false;

	if (map.lines[l]->side1 != -1)
	{
		if (map.l_needsmidtex(l, 1))
		{
			setup_wall_rect(l, 1, 0);
			setup_wall_tex(l, 1, 0);
		}

		if (map.l_needsuptex(l, 1))
		{
			setup_wall_rect(l, 1, 1);
			setup_wall_tex(l, 1, 1);
		}

		if (map.l_needslotex(l, 1))
		{
			setup_wall_rect(l, 1, 2);
			setup_wall_tex(l, 1, 2);
		}
	}

	if (map.lines[l]->side2 != -1)
	{
		if (map.l_needsmidtex(l, 2))
		{
			setup_wall_rect(l, 2, 0);
			setup_wall_tex(l, 2, 0);
		}

		if (map.l_needsuptex(l, 2))
		{
			setup_wall_rect(l, 2, 1);
			setup_wall_tex(l, 2, 1);
		}

		if (map.l_needslotex(l, 2))
		{
			setup_wall_rect(l, 2, 2);
			setup_wall_tex(l, 2, 2);
		}
	}
}

// update_wall: Updates a wall
// ------------------------ >>
void update_wall(int l)
{
	walls[l].has_mid[0] = false;
	walls[l].has_mid[1] = false;
	walls[l].has_up[0] = false;
	walls[l].has_up[1] = false;
	walls[l].has_lo[0] = false;
	walls[l].has_lo[1] = false;

	if (map.lines[l]->side1 != -1)
	{
		if (map.l_needsmidtex(l, 1))
		{
			setup_wall_rect(l, 1, 0);
			setup_wall_tex(l, 1, 0);
		}

		if (map.l_needsuptex(l, 1))
		{
			setup_wall_rect(l, 1, 1);
			setup_wall_tex(l, 1, 1);
		}

		if (map.l_needslotex(l, 1))
		{
			setup_wall_rect(l, 1, 2);
			setup_wall_tex(l, 1, 2);
		}
	}

	if (map.lines[l]->side2 != -1)
	{
		if (map.l_needsmidtex(l, 2))
		{
			setup_wall_rect(l, 2, 0);
			setup_wall_tex(l, 2, 0);
		}

		if (map.l_needsuptex(l, 2))
		{
			setup_wall_rect(l, 2, 1);
			setup_wall_tex(l, 2, 1);
		}

		if (map.l_needslotex(l, 2))
		{
			setup_wall_rect(l, 2, 2);
			setup_wall_tex(l, 2, 2);
		}
	}
}

// update_wall_tex: Updates a wall's textures
// --------------------------------------- >>
void update_wall_tex(string texture)
{
	int side = 0;
	int part = 0;

	if (&walls[h_line].mid[0] == selected_rect)
	{
		side = 1;
		part = 0;
		map.l_setmidtex(h_line, 1, texture);
	}
	else if (&walls[h_line].mid[1] == selected_rect)
	{
		side = 2;
		part = 0;
		map.l_setmidtex(h_line, 2, texture);
	}
	else if (&walls[h_line].up[0] == selected_rect)
	{
		side = 1;
		part = 1;
		map.l_setuptex(h_line, 1, texture);
	}
	else if (&walls[h_line].up[1] == selected_rect)
	{
		side = 2;
		part = 1;
		map.l_setuptex(h_line, 2, texture);
	}
	else if (&walls[h_line].lo[0] == selected_rect)
	{
		side = 1;
		part = 2;
		map.l_setlotex(h_line, 1, texture);
	}
	else if (&walls[h_line].lo[1] == selected_rect)
	{
		side = 2;
		part = 2;
		map.l_setlotex(h_line, 2, texture);
	}

	setup_wall_rect(h_line, side, part);
	setup_wall_tex(h_line, side, part);
}

// update_room_tex: Paints a 'room' with a texture, and updates affected walls
// ------------------------------------------------------------------------ >>
void update_room_tex(string texture)
{
	int side = 0;
	int part = 0;
	string old_tex = tex_list.get_name(selected_rect->tex);

	//if (old_tex.compare("_notex"))
		//old_tex = "-";

	if (&walls[h_line].mid[0] == selected_rect)
	{
		side = 1;
		part = 0;
		map.l_setmidtex(h_line, 1, texture);
	}
	else if (&walls[h_line].mid[1] == selected_rect)
	{
		side = 2;
		part = 0;
		map.l_setmidtex(h_line, 2, texture);
	}
	else if (&walls[h_line].up[0] == selected_rect)
	{
		side = 1;
		part = 1;
		map.l_setuptex(h_line, 1, texture);
	}
	else if (&walls[h_line].up[1] == selected_rect)
	{
		side = 2;
		part = 1;
		map.l_setuptex(h_line, 2, texture);
	}
	else if (&walls[h_line].lo[0] == selected_rect)
	{
		side = 1;
		part = 2;
		map.l_setlotex(h_line, 1, texture);
	}
	else if (&walls[h_line].lo[1] == selected_rect)
	{
		side = 2;
		part = 2;
		map.l_setlotex(h_line, 2, texture);
	}
	
	if (side != 0)
		paint_textures(h_line, side, old_tex, texture);

	for (DWORD a = 0; a < aligned_lines.n_numbers; a++)
		update_wall(aligned_lines.numbers[a]);

	clear_align_lines();
}

// update_sector_flats: Updates any subsectors in a sector
// ---------------------------------------------------- >>
void update_sector_flats(int sector)
{
	setup_sector(sector);

	for (DWORD s = 0; s < n_gl_ssectors; s++)
	{
		if (sects[s]->parent_sector->sector == sector)
			setup_ssect(s);
	}
}

// paste_3d_texture: Pastes the currently copied texture to the hilighted rect
// ------------------------------------------------------------------------ >>
void paste_3d_texture(bool paint_room)
{
	if (h_line != -1)
	{
		if (paint_room)
			update_room_tex(copy_wtex);
		else
			update_wall_tex(copy_wtex);
	}
	else if (h_thing != -1)
	{
		map.things[h_thing]->type = copy_ttype;
		setup_sprite(h_thing);
		update_map_things();
	}
	else if (copy_ftex != "-")
	{
		if (selected_floor != -1)
		{
			map.sectors[selected_floor]->f_tex = copy_ftex;
			update_sector(selected_floor);
		}
		if (selected_ceil != -1)
		{
			map.sectors[selected_ceil]->c_tex = copy_ftex;
			update_sector(selected_ceil);
		}
	}
}

// set_3d_texture: Sets the hilighted rect's texture to the browser texture
// --------------------------------------------------------------------- >>
void set_3d_texture(bool paint_room)
{
	if (h_line != -1)
	{
		if (paint_room)
			update_room_tex(browser_gettex());
		else
			update_wall_tex(browser_gettex());
	}
	else if (h_thing != -1)
	{
		map.things[h_thing]->type = browser_getthing();
		setup_sprite(h_thing);
		update_map_things();
	}
	else
	{
		if (selected_floor != -1)
		{
			map.sectors[selected_floor]->f_tex = browser_gettex();
			update_sector(selected_floor);
		}
		if (selected_ceil != -1)
		{
			map.sectors[selected_ceil]->c_tex = browser_gettex();
			update_sector(selected_ceil);
		}
	}
}

// change_3d_sector_light: Changes the hilighted sector's light level
// --------------------------------------------------------------- >>
void change_3d_sector_light(int amount)
{
	int sector = -1;

	if (h_line == -1)
	{
		if (selected_floor != -1)
			sector = selected_floor;
		if (selected_ceil != -1)
			sector = selected_ceil;
	}
	else if (h_line != -1)
	{
		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
			sector = map.l_getsector1(h_line);
		else
			sector = map.l_getsector2(h_line);
	}

	if (sector == -1)
		return;
	else
	{
		int light = map.sectors[sector]->light + amount;

		if (light == 239)	// Original light was 255, so only subtract 15 to keep 
			light = 240;	// the light level divisible by 16
		if (light > 255)
			light = 255;
		if (light < 0)
			light = 0;

		map.sectors[sector]->light = light;
		update_sector_flats(sector);
		char temp[5] = "";
		sprintf(temp, "%d", light);
		msg_text = "Light Level: ";
		msg_text += temp;
		msg_count = 100;
	}
}

// change_3d_texture_offset: Changes the hilighted wall's texture offset
// ------------------------------------------------------------------ >>
void change_3d_texture_offset(int x_amount, int y_amount)
{
	if (h_line != -1)
	{
		int side = -1;

		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
			side = map.lines[h_line]->side1;
		else
			side = map.lines[h_line]->side2;

		if (side != -1)
		{
			map.sides[side]->x_offset += x_amount;
			map.sides[side]->y_offset += y_amount;

			if (x_amount != 0)
			{
				char temp[8];
				sprintf(temp, "%d", map.sides[side]->x_offset);
				msg_text = "X Offset: ";
				msg_text += temp;
				msg_count = 100;
			}

			if (y_amount != 0)
			{
				char temp[8];
				sprintf(temp, "%d", map.sides[side]->y_offset);
				msg_text = "Y Offset: ";
				msg_text += temp;
				msg_count = 100;
			}
		}

		update_selected_rect();
	}
}

// toggle_3d_texpeg: Toggles upper/lower unpegged on the hilighted wall
// ----------------------------------------------------------------- >>
void toggle_3d_texpeg(bool upper)
{
	if (h_line == -1)
		return;

	if (upper)
	{
		if (map.lines[h_line]->toggle_flag(LINE_UPPERUNPEGGED))
			msg_text = "Upper Unpegged: On";
		else
			msg_text = "Upper Unpegged: Off";

		msg_count = 100;
	}
	else
	{
		if (map.lines[h_line]->toggle_flag(LINE_LOWERUNPEGGED))
			msg_text = "Lower Unpegged: On";
		else
			msg_text = "Lower Unpegged: Off";

		msg_count = 100;
	}

	update_selected_rect();
}

// delete_3d_texture: Changes the hilighted texture to '-'
// ---------------------------------------------------- >>
void delete_3d_texture()
{
	if (h_line != -1)
		update_wall_tex("-");
}

// reset_3d_tex_offsets: Resets the hilighted wall's offsets to 0
// ----------------------------------------------------------- >>
void reset_3d_tex_offsets()
{
	if (h_line != -1)
	{
		int side = -1;

		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
			side = map.lines[h_line]->side1;
		else
			side = map.lines[h_line]->side2;

		if (side != -1)
		{
			map.sides[side]->x_offset = 0;
			map.sides[side]->y_offset = 0;
		}

		update_selected_rect();
	}
}

// align_3d_tex: Aligns textures on the x axis starting at the hilighted wall
// ----------------------------------------------------------------------- >>
void align_3d_tex()
{
	if (h_line != -1)
	{
		int side;

		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
			side = 1;
		else
			side = 2;

		align_textures_x(h_line, side, map.l_getside(h_line, side)->x_offset, tex_list.get_name(selected_rect->tex));

		for (DWORD a = 0; a < aligned_lines.n_numbers; a++)
			update_wall(aligned_lines.numbers[a]);

		clear_align_lines();
	}
}

// change_3d_thingangle: Changes the angle of the hilighted thing
// ----------------------------------------------------------- >>
bool change_3d_thingangle(int amount)
{
	if (h_thing != -1)
	{
		map.things[h_thing]->increment_angle(amount);
		return true;
	}

	return false;
}

// change_3d_thing_z: Changes the z height of the hilighted thing
// ----------------------------------------------------------- >>
bool change_3d_thing_z(int amount)
{
	if (h_thing != -1 && map.hexen)
	{
		map.things[h_thing]->z += amount;
		char temp[8];
		sprintf(temp, "%d", map.things[h_thing]->z);
		msg_text = "Thing Height: ";
		msg_text += temp;
		msg_count = 100;
		return true;
	}

	return false;
}
