// << ------------------------------------ >>
// << SLADE - SlayeR's LeetAss Doom Editor >>
// << By Simon Judd, 2004                  >>
// << ------------------------------------ >>
// << draw.cpp - Basic drawing functions   >>
// << ------------------------------------ >>

// Disable conversion warnings
#pragma warning (disable: 4244)

// INCLUDES ------------------------------ >>
#include "main.h"
#include "gui.h"

// VARIABLES ----------------------------- >>
char charmap[16][16] =
{
	{ 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P'},
	{ 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f'},
	{ 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v'},
	{ 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '(', ')'},
	{ '!', '?', '#', '&', '*', '.', ',', ':', ';', '"', '\'', '/', '-', '+', '_', '%'},
	{ '=', '{', '}', '<', '>', '[', ']', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
	{ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '}
};

// EXTERNAL VARIABLES -------------------- >>
extern TextureList tex_list;
extern float x_scale, y_scale;

EXTERN_CVAR(Bool, gui_scale)

// set_colour: Sets the current OpenGL colour
// ------------------------------------------ >>
void set_colour(rgba_t col)
{
	glColor4f(col.fr(), col.fg(), col.fb(), col.fa());
}

// draw_point: Draws a point of given size at (x, y)
// ---------------------------------------------- >>
void draw_point(int x, int y, int size, rgba_t col)
{
	glDisable(GL_TEXTURE_2D);

	set_colour(col);
	glPointSize((float)size);

	glBegin(GL_POINTS);
		glVertex2d(x, y);
	glEnd();
}

// draw_line: Draws a line from a rect's top left to it's bottom right
// ---------------------------------------------------------------- >>
void draw_line(rect_t rect, rgba_t col, bool aa)
{
	glDisable(GL_TEXTURE_2D);

	if (aa)
		glEnable(GL_LINE_SMOOTH);
	else
		glDisable(GL_LINE_SMOOTH);

	set_colour(col);
	
	glBegin(GL_LINES);
		glVertex2d(rect.tl.x, rect.tl.y);
		glVertex2d(rect.br.x, rect.br.y);
	glEnd();
}

// draw_rect: Draws a rectangle
// ------------------------- >>
void draw_rect(rect_t rect, rgba_t col, bool fill)
{
	glDisable(GL_TEXTURE_2D);

	set_colour(col);

	if (fill)
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	
	glBegin(GL_QUADS);
	glVertex2d(rect.x1(), rect.y1());
	glVertex2d(rect.x2(), rect.y1());
	glVertex2d(rect.x2(), rect.y2());
	glVertex2d(rect.x1(), rect.y2());
	glEnd();

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

// draw_text: Draws a string of text at x, y (characters are 8x8)
// ----------------------------------------------------------- >>
void draw_text(int x, int y, rgba_t colour, BYTE alignment, const char* text, ...)
{
	float tx1, tx2, ty1, ty2;
	char string[256];
	va_list ap;

	va_start(ap, text);
	vsprintf(string, text, ap);
	va_end(ap);

	float size_x = 8 * x_scale;
	float size_y = 8 * y_scale;

	// Alignment
	if (alignment == 1) // Center
		x -= (((int)strlen(string) * size_x) / 2);

	if (alignment == 2) // Right
		x -= ((int)strlen(string) * size_x);

	set_colour(colour);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glBindTexture(GL_TEXTURE_2D, tex_list.gl_tex("_font"));

	for (DWORD char_num = 0; char_num < strlen(string); char_num++)
	{
		for (int row = 0; row < 16; row++)
		{
			for (int column = 0; column < 16; column++)
			{
				if (string[char_num] == charmap[row][column])
				{
					tx1 = (float)column * 0.0625f;
					tx2 = (float)(column + 1) * 0.0625f;
					ty1 = (float)row * 0.0625f;
					ty2 = (float)(row + 1) * 0.0625f;

					glBegin(GL_QUADS);
						glTexCoord2f(tx1, ty1); glVertex2d((char_num * size_x) + x, y);
						glTexCoord2f(tx2, ty1); glVertex2d((char_num * size_x) + x + size_x, y);
						glTexCoord2f(tx2, ty2); glVertex2d((char_num * size_x) + x + size_x, y + size_y);
						glTexCoord2f(tx1, ty2); glVertex2d((char_num * size_x) + x, y + size_y);
					glEnd();
				}
			}
		}
	}
}

// draw_texture: Draws a texture at x, y, resized to the given width & height
// not resized if width or height = -1
// ----------------------------------------------------------------------- >>
void draw_texture(string name, int x, int y, int width, int height, rgba_t col)
{
	texture_t *tex = tex_list.get_tex(name);

	if (width == -1)
		width = tex->width;

	if (height == -1)
		height = tex->height;

	/*
	if (gui_scale)
	{
		x *= x_scale;
		y *= y_scale;
		width *= x_scale;
		height *= y_scale;
	}
	*/

	set_colour(col);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
	glBindTexture(GL_TEXTURE_2D, tex->tex_id);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f);				glVertex2d(x, y);
	glTexCoord2f(tex->tex_x, 0.0f);			glVertex2d(x + width, y);
	glTexCoord2f(tex->tex_x, tex->tex_y);	glVertex2d(x + width, y + height);
	glTexCoord2f(0.0f, tex->tex_y);			glVertex2d(x, y + height);
	glEnd();
}

// draw_texture: Draws a texture at x, y, resized to the given width & height
// not resized if width or height = -1
// ----------------------------------------------------------------------- >>
void draw_texture(texture_t *tex, int x, int y, int width, int height, rgba_t col)
{
	if (width == -1)
		width = tex->width;

	if (height == -1)
		height = tex->height;

	set_colour(col);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
	glBindTexture(GL_TEXTURE_2D, tex->tex_id);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f);				glVertex2d(x, y);
	glTexCoord2f(tex->tex_x, 0.0f);			glVertex2d(x + width, y);
	glTexCoord2f(tex->tex_x, tex->tex_y);	glVertex2d(x + width, y + height);
	glTexCoord2f(0.0f, tex->tex_y);			glVertex2d(x, y + height);
	glEnd();
}

// draw_texture: Draws a texture at x, y, resized to the given width & height
// not resized if width or height = -1
// ----------------------------------------------------------------------- >>
void draw_texture(string name, int x, int y, int width, int height, rgba_t col, bool available, bool flat)
{
	texture_t *tex = tex_list.get_tex(name, available, flat);

	if (width == -1)
		width = tex->width;

	if (height == -1)
		height = tex->height;

	set_colour(col);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
	glBindTexture(GL_TEXTURE_2D, tex->tex_id);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f);				glVertex2d(x, y);
	glTexCoord2f(tex->tex_x, 0.0f);			glVertex2d(x + width, y);
	glTexCoord2f(tex->tex_x, tex->tex_y);	glVertex2d(x + width, y + height);
	glTexCoord2f(0.0f, tex->tex_y);			glVertex2d(x, y + height);
	glEnd();
}

// draw_texture_scale: Draws a texture at x, y, resized to fit within the given width & height
// BUT keeps the correct scale
// ----------------------------------------------------------------------- >>
void draw_texture_scale(string name, int x, int y, int width, int height, rgba_t col, bool available, bool flat)
{
	int mid_x = x + (width / 2);
	int mid_y = y + (height / 2);

	int nx1 = x;
	int nx2 = x + width;
	int ny1 = y;
	int ny2 = y + height;
	
	texture_t *tex = tex_list.get_tex(name, available, flat);
	
	if (!tex)
	{
		printf("Texture %s not found.\n", name.c_str());
		return;
	}
	
	/*
	if (tex->width < tex->height)
	{
		int mid = x + (width / 2);
		float multiplier = (float)height / (float)tex->height;
		
		nx1 = mid - (float(tex->width / 2.0f) * multiplier);
		nx2 = mid + (float(tex->width / 2.0f) * multiplier);
	}
	
	if (tex->width > tex->height)
	{
		int mid = y + (height / 2);
		float multiplier = (float)width / (float)tex->width;
		
		ny1 = mid - (float(tex->height / 2.0f) * multiplier);
		ny2 = mid + (float(tex->height / 2.0f) * multiplier);
	}
	*/

	if (tex->width > width)
	{
		float multiplier = (float)width / (float)tex->width;
		
		ny1 = mid_y - (float(tex->height / 2.0f) * multiplier);
		ny2 = mid_y + (float(tex->height / 2.0f) * multiplier);
		height = tex->height * multiplier;
	}
	else
	{
		nx1 = mid_x - (tex->width / 2);
		nx2 = mid_x + (tex->width / 2);
	}

	if (tex->height > height)
	{
		float multiplier = (float)height / (float)tex->height;
		
		nx1 = mid_x - (float(tex->width / 2.0f) * multiplier);
		nx2 = mid_x + (float(tex->width / 2.0f) * multiplier);
	}
	else
	{
		ny1 = mid_y - (tex->height / 2);
		ny2 = mid_y + (tex->height / 2);
	}
	
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex->tex_id);
	
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f);				glVertex2d(nx1, ny1);
		glTexCoord2f(tex->tex_x, 0.0f);			glVertex2d(nx2, ny1);
		glTexCoord2f(tex->tex_x, tex->tex_y);	glVertex2d(nx2, ny2);
		glTexCoord2f(0.0f, tex->tex_y);			glVertex2d(nx1, ny2);
	glEnd();
}

// draw_texture_scale: Draws a texture at x, y, resized to fit within the given width & height
// BUT keeps the correct scale
// ----------------------------------------------------------------------- >>
void draw_texture_scale(texture_t *tex, int x, int y, int width, int height, rgba_t col, bool available, bool flat)
{
	int mid_x = x + (width / 2);
	int mid_y = y + (height / 2);

	int nx1 = x;
	int nx2 = x + width;
	int ny1 = y;
	int ny2 = y + height;
	
	//texture_t *tex = tex_list.get_tex(name, available, flat);
	
	if (!tex)
	{
		//printf("Texture %s not found.\n", name.c_str());
		return;
	}
	
	/*
	if (tex->width < tex->height)
	{
		int mid = x + (width / 2);
		float multiplier = (float)height / (float)tex->height;
		
		nx1 = mid - (float(tex->width / 2.0f) * multiplier);
		nx2 = mid + (float(tex->width / 2.0f) * multiplier);
	}
	
	if (tex->width > tex->height)
	{
		int mid = y + (height / 2);
		float multiplier = (float)width / (float)tex->width;
		
		ny1 = mid - (float(tex->height / 2.0f) * multiplier);
		ny2 = mid + (float(tex->height / 2.0f) * multiplier);
	}
	*/

	if (tex->width > width)
	{
		float multiplier = (float)width / (float)tex->width;
		
		ny1 = mid_y - (float(tex->height / 2.0f) * multiplier);
		ny2 = mid_y + (float(tex->height / 2.0f) * multiplier);
		height = tex->height * multiplier;
	}
	else
	{
		nx1 = mid_x - (tex->width / 2);
		nx2 = mid_x + (tex->width / 2);
	}

	if (tex->height > height)
	{
		float multiplier = (float)height / (float)tex->height;
		
		nx1 = mid_x - (float(tex->width / 2.0f) * multiplier);
		nx2 = mid_x + (float(tex->width / 2.0f) * multiplier);
	}
	else
	{
		ny1 = mid_y - (tex->height / 2);
		ny2 = mid_y + (tex->height / 2);
	}
	
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex->tex_id);
	
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f);				glVertex2d(nx1, ny1);
		glTexCoord2f(tex->tex_x, 0.0f);			glVertex2d(nx2, ny1);
		glTexCoord2f(tex->tex_x, tex->tex_y);	glVertex2d(nx2, ny2);
		glTexCoord2f(0.0f, tex->tex_y);			glVertex2d(nx1, ny2);
	glEnd();
}

// draw_texture_rect: Draws part of a texture at x, y
// ----------------------------------------------- >>
void draw_texture_rect(string name, int x, int y, int width, int height, rect_t part, rgba_t col)
{
	texture_t *tex = tex_list.get_tex(name);

	float tx1 = tex->tex_x * ((float)part.x1() / (float)tex->width);
	float ty1 = tex->tex_y * ((float)part.y1() / (float)tex->height);
	float tx2 = tex->tex_x * ((float)part.x2() / (float)tex->width);
	float ty2 = tex->tex_y * ((float)part.y2() / (float)tex->height);

	/*
	if (gui_scale)
	{
		x *= x_scale;
		y *= y_scale;
		width *= x_scale;
		height *= y_scale;
	}
	*/

	set_colour(col);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
	glBindTexture(GL_TEXTURE_2D, tex->tex_id);

	glBegin(GL_QUADS);
	glTexCoord2f(tx1, ty1);	glVertex2d(x, y);
	glTexCoord2f(tx2, ty1);	glVertex2d(x + width, y);
	glTexCoord2f(tx2, ty2);	glVertex2d(x + width, y + height);
	glTexCoord2f(tx1, ty2);	glVertex2d(x, y + height);
	glEnd();
}
