// << ------------------------------------ >>
// << SLADE - SlayeR's LeetAss Doom Editor >>
// << By Simon Judd, 2004-05               >>
// << ------------------------------------ >>
// << edit.cpp - Various editing tools     >>
// << (for the 'edit' menu)                >>
// << ------------------------------------ >>

// INCLUDES ------------------------------ >>
#include "main.h"
#include "map.h"

// VARIABLES ----------------------------- >>

// EXTERNAL VARIABLES -------------------- >>
extern Map map;
extern numlist_t selected_items;

// sector_create_stairs: Creates stairs from a group of sectors
// --------------------------------------------------------- >>
void sector_create_stairs(int step, bool ceil)
{
	if (selected_items.n_numbers <= 1)
		return;

	int height = map.sectors[selected_items.numbers[0]]->f_height;

	for (DWORD a = 1; a < selected_items.n_numbers; a++)
	{
		height += step;
		map.sectors[selected_items.numbers[a]]->f_height = height;
	}

	if (ceil)
	{
		int height = map.sectors[selected_items.numbers[0]]->c_height;

		for (DWORD a = 1; a < selected_items.n_numbers; a++)
		{
			height += step;
			map.sectors[selected_items.numbers[a]]->c_height = height;
		}
	}

	map_changelevel(2);
}

// sector_create_door: Creates a door from any selected sectors
// --------------------------------------------------------- >>
void sector_create_door(string texture)
{
	if (selected_items.n_numbers == 0)
		return;

	for (DWORD a = 0; a < selected_items.n_numbers; a++)
	{
		int s = selected_items.numbers[a];

		// Set ceiling to floor
		map.sectors[s]->c_height = map.sectors[s]->f_height;
		
		// Go through lines in the sector, set their texture and type accordingly
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (map.lines[l]->side2 != -1)
			{
				// Any 2-sided lines
				if (map.sides[map.lines[l]->side2]->sector == s)
				{
					map.sides[map.lines[l]->side1]->tex_upper = texture;
					
					if (!map.zdoom)
						map.lines[l]->type = 1;
					else
					{
						map.lines[l]->type = 12;
						map.lines[l]->args[1] = 24;
						map.lines[l]->args[2] = 128;
					}
				}

				if (map.sides[map.lines[l]->side1]->sector == s)
				{
					map.sides[map.lines[l]->side2]->tex_upper = texture;

					if (!map.zdoom)
						map.lines[l]->type = 1;
					else
					{
						map.lines[l]->type = 12;
						map.lines[l]->args[1] = 24;
						map.lines[l]->args[2] = 128;
					}
				}
			}
			else // 1-sided lines
			{
				if (map.sides[map.lines[l]->side1]->sector == s)
				{
					map.sides[map.lines[l]->side1]->tex_middle = "DOORTRAK";
					map.lines[l]->set_flag(LINE_LOWERUNPEGGED);
				}
			}
		}
	}

	map_changelevel(2);
}
