#define	LINE_IMPASSIBLE		0x0001
#define LINE_BLOCKMONSTERS	0x0002
#define	LINE_TWOSIDED		0x0004
#define	LINE_UPPERUNPEGGED	0x0008
#define	LINE_LOWERUNPEGGED	0x0010
#define	LINE_SECRET			0x0020
#define	LINE_BLOCKSOUND		0x0040
#define	LINE_NOTONMAP		0x0080
#define	LINE_STARTONMAP		0x0100
#define LINE_REPEATABLE		0x0200
#define LINE_MONSTERCANACT	0x2000
#define LINE_BLOCKALL		0x8000

#define LINE_SPAC_CROSS			0
#define LINE_SPAC_USE			1
#define LINE_SPAC_MCROSS		2
#define LINE_SPAC_IMPACT		3
#define LINE_SPAC_PUSH			4
#define LINE_SPAC_PCROSS		5
#define LINE_SPAC_USETHROUGH	6
#define LINE_SPAC_PTOUCH		7

struct linedef_t
{
	// Common info
	WORD	vertex1;
	WORD	vertex2;
	WORD	flags;
	WORD	type;
	short	side1;
	short	side2;

	// Doom format only
	short	sector_tag;

	// Hexen format only
	BYTE	args[5];

	// Editor stuff
	/*
	bool	hilighted;
	bool	selected;
	bool	moving;
	*/

	linedef_t();

	linedef_t(int V1, int V2)
	{
		vertex1 = V1;
		vertex2 = V2;

		flags = 0;
		type = 0;
		side1 = -1;
		side2 = -1;
		sector_tag = 0;
	
		for (int a = 0; a < 5; a++)
			args[a] = 0;

		//hilighted = selected = moving = false;
	}

	/*
	void toggle_select()
	{
		selected = !selected;
	}

	void select()
	{
		selected = true;
	}

	void deselect()
	{
		selected = false;
	}
	*/

	bool toggle_flag(WORD flag)
	{
		flags ^= flag;

		return !!(flags & flag);
	}

	void set_flag(WORD flag)
	{
		flags |= flag;
	}

	void clear_flag(WORD flag)
	{
		flags = (flags & ~flag);
	}

	WORD get_trigger()
	{
		return ((flags & 0x1c00) >> 10);
	}

	void set_trigger(WORD trigger)
	{
		//if (trigger > 0)
			//trigger = (trigger - 1) << 10;

		trigger = trigger << 10;

		flags &= ~0x1c00;
		flags |= trigger;
	}

	string get_trigger_string()
	{
		WORD trigger = get_trigger();

		if (trigger == 0)
			return "Player Crosses Line";
		if (trigger == 1)
			return "Player Uses Line";
		if (trigger == 2)
			return "Monster Crosses Line";
		if (trigger == 3)
			return "Projectile Hits Line";
		if (trigger == 4)
			return "Player Bumps Line";
		if (trigger == 5)
			return "Projectile Crosses Line";
		if (trigger == 6)
			return "Player Uses Line (PassThru)";
		if (trigger == 7)
			return "Projectile Hits OR Crosses Line";

		return "Unknown";
	}
};

struct linetype_t
{
	BYTE	type;
	string	name;
	char	arg_tag;	// -1 = doom format tagged, 0 = no tag, 1-5 = args
	string	args[5];
	string*	arg_types[5];
	
	linetype_t();

	void dump()
	{
		printf("Line %d: \"%s\", %d\n", type, name.c_str(), arg_tag);
	}
};

struct linegroup_t
{
	string		name;
	linetype_t*	*types;
	DWORD		n_types;

	linegroup_t()
	{
		name = "-";
		types = NULL;
		n_types = 0;
	}

	void add(linetype_t *type)
	{
		n_types++;
		types = (linetype_t **)realloc(types, n_types * sizeof(linetype_t *));
		types[n_types - 1] = type;
	}
};

struct linelist_t
{
	linedef_t*	*lines;
	DWORD		n_lines;

	linelist_t()
	{
		lines = NULL;
		n_lines = 0;
	}

	void add(linedef_t* line)
	{
		n_lines++;
		lines = (linedef_t **)realloc(lines, n_lines * sizeof(linedef_t *));
		lines[n_lines - 1] = line;
	}

	linedef_t* last() { return lines[n_lines - 1]; }
	linedef_t* first() { return lines[0]; }
};

void parse_line_group(MemReader *mr);
linetype_t* get_ltype(BYTE type);
void line_settype(BYTE group, WORD item, DWORD line);
void thing_setspecial(BYTE group, WORD item, DWORD thing);
