
#include "main.h"
#include "map.h"

numlist_t aligned_lines;

extern Map map;
extern TextureList tex_list;

// align_textures_x: Recursively aligns textures starting from a single line
// ---------------------------------------------------------------------- >>
void align_textures_x(int start_line, int side, int offset, string texture)
{
	// Don't align a blank texture :P
	if (texture.compare("-") == 0)
		return;

	int tex_width = tex_list.get_tex(texture, true, false)->width;
	bool done = false;

	// If we have an offset greater than the texture width, adjust it
	while (!done)
	{
		done = true;

		if (offset > tex_width)
		{
			offset -= tex_width;
			done = false;
		}
	}

	// If we have an offset less than 0, adjust it
	done = false;
	while (!done)
	{
		done = true;

		if (offset < 0)
		{
			offset += tex_width;
			done = false;
		}
	}

	// If we have an offset equal to the texture width, it should be just 0
	if (offset == tex_width)
		offset = 0;

	// Add the line to a list, and set the offset
	aligned_lines.add(start_line, true);
	map.l_getside(start_line, 1)->x_offset = offset;

	rect_t line = map.l_getrect(start_line);
	int length = (int)distance(line.x1(), line.y1(), line.x2(), line.y2());
	int vertex = 0;
	numlist_t a_lines;

	if (side == 1)
		vertex = map.lines[start_line]->vertex2;
	else
		vertex = map.lines[start_line]->vertex1;

	// Go through all attached lines that have the same texture
	map.v_getattachedlines(vertex, &a_lines);

	for (DWORD a = 0; a < a_lines.n_numbers; a++)
	{
		DWORD l = a_lines.numbers[a];

		if (map.l_getside(l, 1) && !aligned_lines.exists(l))
		{
			if (map.l_getside(l, 1)->tex_middle == texture
				|| map.l_getside(l, 1)->tex_upper == texture
				|| map.l_getside(l, 1)->tex_lower == texture)
				align_textures_x(l, 1, offset + length, texture);
		}

		if (map.l_getsector2(l) != -1 && !aligned_lines.exists(l))
		{
			if (map.l_getside(l, 2)->tex_middle == texture
				|| map.l_getside(l, 2)->tex_upper == texture
				|| map.l_getside(l, 2)->tex_lower == texture)
				align_textures_x(l, 2, offset + length, texture);
		}
	}
}

bool line_hastex(int line, int side, string texture)
{
	if (side == 1)
	{
		if (map.sides[map.lines[line]->side1]->tex_middle == texture
			&& map.l_needsmidtex(line, side))
			return true;

		if (map.sides[map.lines[line]->side1]->tex_upper == texture
			&& map.l_needsuptex(line, side))
			return true;

		if (map.sides[map.lines[line]->side1]->tex_lower == texture
			&& map.l_needslotex(line, side))
			return true;

		return false;
	}
	else
	{
		if (map.sides[map.lines[line]->side2]->tex_middle == texture
			&& map.l_needsmidtex(line, side))
			return true;

		if (map.sides[map.lines[line]->side2]->tex_upper == texture
			&& map.l_needsuptex(line, side))
			return true;

		if (map.sides[map.lines[line]->side2]->tex_lower == texture
			&& map.l_needslotex(line, side))
			return true;

		return false;
	}
}

// paint_textures: 'Paints' over all textures matching 'oldtex' with 'newtex'
// (probably doesn't belong here but oh well)
// ----------------------------------------------------------------------- >>
void paint_textures(int start_line, int side, string oldtex, string newtex)
{
	if (aligned_lines.exists(start_line))
		return;

	if (oldtex == "_notex")
		oldtex = "-";

	// Add the line to a list, and set the texture
	aligned_lines.add(start_line, true);

	if (map.l_getside(start_line, side)->tex_middle == oldtex
		&& map.l_needsmidtex(start_line, side))
		map.l_getside(start_line, side)->tex_middle = newtex;

	if (map.l_getside(start_line, side)->tex_upper == oldtex
		&& map.l_needsuptex(start_line, side))
		map.l_getside(start_line, side)->tex_upper = newtex;

	if (map.l_getside(start_line, side)->tex_lower == oldtex
		&& map.l_needslotex(start_line, side))
		map.l_getside(start_line, side)->tex_lower = newtex;

	int vertex = map.lines[start_line]->vertex1;
	numlist_t a_lines;

	// Go through all attached lines that have the same texture
	map.v_getattachedlines(vertex, &a_lines);

	for (DWORD a = 0; a < a_lines.n_numbers; a++)
	{
		int l = a_lines.numbers[a];

		if (map.l_getside(l, 1))
		{
			if (line_hastex(l, 1, oldtex))
				paint_textures(l, 1, oldtex, newtex);
		}

		if (map.l_getside(l, 2))
		{
			if (line_hastex(l, 2, oldtex))
				paint_textures(l, 2, oldtex, newtex);
		}
	}

	// Go again but this time from the line's second vertex
	vertex = map.lines[start_line]->vertex2;
	map.v_getattachedlines(vertex, &a_lines);

	for (DWORD a = 0; a < a_lines.n_numbers; a++)
	{
		int l = a_lines.numbers[a];

		if (map.l_getside(l, 1))
		{
			if (line_hastex(l, 1, oldtex))
				paint_textures(l, 1, oldtex, newtex);
		}

		if (map.l_getside(l, 2))
		{
			if (line_hastex(l, 2, oldtex))
				paint_textures(l, 2, oldtex, newtex);
		}
	}
}

void clear_align_lines()
{
	aligned_lines.clear();
}
